/* o_dup.c - Generic redundancy - Don Yang / Dave Newquist / John Rapp

   07/28/01
                                                                           */

#include<assert.h>
#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include"smlng.h"
#include"token.h"
#include"tree.h"


/* Local functions */
static void MergeNeighbors(Node *node);
static int RedundantAttribute(Node *node);
static int SameTag(Token *a, Token *b);

/************************************************************* o_duplicate */
int o_duplicate(Node **node)
{
   int modified;

   modified = 0;
   while( *node )
   {
      if( (*node)->token->tagtype )
      {
         /* Tags */
         if( (*node)->next != NULL &&
             SameTag((*node)->token, (*node)->next->token) )
         {
            MergeNeighbors(*node);
            modified = 1;
            continue;
         }
         else if( RedundantAttribute(*node) )
         {
            PromoteSubtree(node);            /* (tree.c) */
            modified = 1;
            continue;
         }
      }

      if( (*node)->tree )
         modified |= o_duplicate(&((*node)->tree));
      node = &((*node)->next);
   }

   return modified;

} /* o_duplicate() */


/********************************************************** MergeNeighbors */
static void MergeNeighbors(Node *node)
{
   Node *tail, *del;

   assert(node->next != NULL);

   del = node->next;
   if( node->tree == NULL )
   {
      /* Current node encloses no range, move neighbor's content to here */
      node->tree = del->tree;
   }
   else
   {
      /* Join head of neighbor's subtree to tail of current node's subtree */
      for(tail = node->tree; tail->next; tail = tail->next);
      tail->next = del->tree;
   }

   /* Delete neighbor node */
   node->next = del->next;
   del->next = del->tree = NULL;
   DeleteParseTree(del);

} /* MergeNeighbors() */

/****************************************************** RedundantAttribute */
static int RedundantAttribute(Node *node)
{
   switch( node->token->type )
   {
      case TagB:     return node->attr.b;
      case TagI:     return node->attr.i;
      case TagS:     return node->attr.s;
      case TagTT:    return node->attr.tt;

      case TagEM:    return node->attr.s;

      case TagSize:  return node->attr.size == node->token->lexeme.i;
      case TagColor:
         return node->attr.color == node->token->lexeme.i;
      default:
         return 0;
   }

} /* RedundantAttribute() */

/***************************************************************** SameTag */
static int SameTag(Token *a, Token *b)
{
   if( a->type != b->type )
      return 0;
   if( a->type == TagSize || a->type == TagColor )
      return a->lexeme.i == b->lexeme.i;
   return 1;

} /* SameTag() */
