/* o_em.c - Reorder EM tags - Don Yang / Dave Newquist / John Rapp

   07/28/01
                                                                           */

#include<assert.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include"smlng.h"
#include"token.h"
#include"tree.h"


/* Local functions */
static int DupEMRange(Node **node);
static int SRange(Node **node);

/******************************************************************** o_em */
int o_em(Node **node)
{
   int modified;

   modified = 0;
   while( *node != NULL )
   {
      if( (*node)->token->type == TagEM )
      {
         /* Remove trivial nested EM ranges */
         if( DupEMRange(node) )
         {
            modified = 1;
            continue;
         }
         else
         {
            /* S range */
            modified |= SRange(node);
         }
      }

      if( (*node)->tree )
         modified |= o_em(&((*node)->tree));
      node = &((*node)->next);
   }
   return modified;

} /* o_em() */


/************************************************************** DupEMRange */
static int DupEMRange(Node **node)
{
   Node *tail, *del, *n;

   if( (del = (*node)->tree) == NULL )
   {
      DeleteNode(node);       /* (tree.c) */
      return 1;
   }
   if( del->token->type != TagEM )
      return 0;

   if( del->tree == NULL )
   {
      /* Second level of EM contains nothing */
      DeleteNode(&((*node)->tree));       /* (tree.c) */
   }
   else
   {
      /* Promote subtree */
      for(tail = del->tree; tail->next; tail = tail->next);
      n = *node;
      *node = del->tree;
      tail->next = n;

      assert(n->tree == del);
      n->tree = del->next;

      del->next = del->tree = NULL;
      DeleteParseTree(del);            /* (tree.c) */
   }

   return 1;

} /* DupEMRange() */

/****************************************************************** SRange */
static int SRange(Node **node)
{
   Node *tail, *em;

   /* Find second EM node */
   for(em = (*node)->next; em && em->token->type != TagEM; em = em->next)
   {
      if( em->token->type != TagS )
         return 0;
   }

   if( em == NULL )
      return 0;

   /* If next EM tag is the adjacent node, leave it and let o_dup
      merge the two tags.                                         */
   if( em == (*node)->next )
      return 0;
   if( (*node)->tree == NULL )
      return 0;

   /* Move enclosed S range to end of first EM subtree */
   for(tail = (*node)->tree; tail->next; tail = tail->next);
   tail->next = (*node)->next;

   /* Move second EM to be enclosed in first EM */
   for(tail = tail->next; tail->next != em; tail = tail->next);
   tail->next = em->tree;
   (*node)->next = em->next;
   em->next = em->tree = NULL;
   DeleteParseTree(em);    /* (tree.c) */

   return 1;

} /* SRange() */
