/* o_spc.c - Whitespace optimizations - Don Yang / Dave Newquist / John Rapp

   07/28/01
                                                                           */

#include<assert.h>
#include<stdio.h>
#include<string.h>
#include"smlng.h"
#include"token.h"
#include"tree.h"


/* Local functions */
static int AppendSpace(Node **node);
static int PropagateSpace(Node **node);
static int SameSpace(Attr *a, Attr *b);

/************************************************************ o_whitespace */
int o_whitespace(Node **node)
{
   int modified, space;

   modified = space = 0;
   if( (*node)->attr.tt )
   {
      /* Typewrite mode, preserve all whitespaces */
      while( *node )
      {
         if( (*node)->tree )
            modified |= o_whitespace(&((*node)->tree));
         node = &((*node)->next);
      }
      return modified;
   }

   /* Remove all consecutive spaces at this level */
   while( *node )
   {
      if( (*node)->token->type == TokenSpace )
      {
         /* Space */
         if( space )
         {
            /* Consolidate */
            DeleteNode(node);    /* (tree.c) */
            modified = 1;
            continue;
         }
         else
         {
            /* Print first space */
            space = 1;
         }
      }
      else
      {
         /* Non-whitespace */
         if( (*node)->tree )
         {
            if( o_whitespace(&((*node)->tree)) )
            {
               modified = 1;
            }
            else
            {
               if( PropagateSpace(node) )
                  modified = 1;
               else
                  modified |= AppendSpace(node);
            }
         }
         space = 0;
      }
      node = &((*node)->next);
   }

   return modified;

} /* o_whitespace() */


/************************************************************* AppendSpace */
static int AppendSpace(Node **node)
{
   Node *space, *tail;

   assert(*node != NULL && (*node)->tree != NULL);
   if( (space = (*node)->next) == NULL )
      return 0;

   /* Shift trailing spaces to inside of range */
   if( space->token->type == TokenSpace )
   {
      for(tail = (*node)->tree; tail->next; tail = tail->next);
      if( SameSpace(&(tail->attr), &(space->attr)) )
      {
         tail->next = space;
         (*node)->next = space->next;
         space->next = NULL;
         memcpy(&(space->attr), &(tail->attr), sizeof(Attr));
         return 1;
      }
   }

   return 0;

} /* AppendSpace() */

/********************************************************** PropagateSpace */
static int PropagateSpace(Node **node)
{
   Node *space, *n;

   assert(*node != NULL && (*node)->tree != NULL);

   /* Shift enclosed leading spaces to outside of range */
   if( TokenSpace == (space = (*node)->tree)->token->type )
   {
      if( SameSpace(&((*node)->attr), &(space->attr)) )
      {
         assert(space->tree == NULL);

         /* Copy attribute */
         memcpy(&(space->attr), &((*node)->attr), sizeof(Attr));

         /* Move node */
         n = space->next;
         space->next = *node;
         (*node)->tree = n;
         *node = space;
         return 1;
      }
   }

   return 0;

} /* PropagateSpace() */

/*************************************************************** SameSpace */
static int SameSpace(Attr *a, Attr *b)
{
   if( a->size != b->size || a->u != b->u || a->tt || b->tt )
      return 0;
   if( a->u == 0 )
      return 1;
   return a->color == b->color;

} /* SameSpace() */
